import torch
import gpytorch
import warnings
import faiss
import numpy as np


class BaseRegressor(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.history = {"train_loss": []}
        self.best_restart = None
        self.res = faiss.StandardGpuResources()

    def compute_train_nn_idx(self, k):
        x = (self.train_x.data.float()).cpu().numpy()
        self.cpu_index = faiss.IndexFlatL2(self.train_x.size(-1))
        self.gpu_index = faiss.index_cpu_to_gpu(self.res, 1, self.cpu_index)
        self.gpu_index.add(x)
        self.train_nn_idx = (
            torch.from_numpy(self.gpu_index.search(x, k)[1])
            .long()
            .to(self.train_x.device)
        )

    def fit(
        self,
        batch_size=None,
        lr=0.1,
        n_epochs=1,
        n_restarts=1,
        verbose=0,
        verbose_gap=1,
        random_state=None,
    ):
        if random_state is not None:
            torch.manual_seed(random_state)
        if batch_size is None:
            batch_size = self.train_x.shape[0]

        self.model.train()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)

        if len(self.history["train_loss"]) > 0:
            warnings.warn(
                "Training loss is not empty. This may cause unexpected behavior."
            )

        self.compute_train_nn_idx(k=batch_size)

        best_loss = float("inf")
        for restart in range(n_restarts):
            self.history["train_loss"].append([])
            for param in self.model.parameters():
                torch.nn.init.normal_(param, mean=0.0, std=1.0)

            for epoch in range(n_epochs):
                for iteration in range(self.train_x.shape[0] // batch_size):
                    idx = torch.randint(low=0, high=self.train_x.shape[0], size=(1,))[0]
                    indices = self.train_nn_idx[idx]
                    X_batch = self.train_x[indices]
                    y_batch = self.train_y[indices]
                    # X_batch = self.train_x
                    # y_batch = self.train_y

                    self.optimizer.zero_grad()
                    loss = self(X_batch, y_batch)
                    loss.backward()
                    if verbose and epoch % verbose_gap == 0:
                        print(
                            "Restart: {}, Iter: {}, Loss: {:.4f}, Best Loss: {:.4f}".format(
                                restart, epoch, loss.item(), best_loss
                            )
                        )
                    self.history["train_loss"][restart].append(loss.item())
                    self.optimizer.step()

            # Last loss
            # This can consume entire RAM
            loss = self(self.train_x, self.train_y)

            # Check if best loss
            if loss.item() < best_loss:
                self.best_restart = restart
                best_loss = loss.item()
                best_model_state = self.model.state_dict()

            # Load the best model
            self.model.load_state_dict(best_model_state)



import os
import pandas as pd

data = pd.read_csv(
    os.path.join(os.path.dirname(__file__), "laptops.csv"), encoding="latin-1"
)
print(data.head())
print(data.columns)


def screenfilter(x):
    splits = x.strip().split()
    if len(splits) > 1:
        return " ".join(splits[:-1])
    else:
        return "Unknown"


def screenresx(x):
    splits = x.strip().split()
    return int(splits[-1].split("x")[0])


def screenresy(x):
    splits = x.strip().split()
    return int(splits[-1].split("x")[1])


data["ScreenType"] = data["ScreenResolution"].apply(screenfilter)
data["ScreenResolutionX"] = data["ScreenResolution"].apply(screenresx)
data["ScreenResolutionY"] = data["ScreenResolution"].apply(screenresy)
for i, j, k in zip(
    data["ScreenResolution"], data["ScreenResolutionX"], data["ScreenResolutionY"]
):
    # print(i, "@@@", j, "x", k)
    pass

data.rename(columns={"ScreenResolution": "OldFeature_ScreenResolution"}, inplace=True)


# print(data["ScreenType"].unique())
# print(data["ScreenResolutionX"].unique())
# print(data["ScreenResolutionY"].unique())
# print(data["Cpu"])


def cputype(x):
    splits = x.strip().split()
    if len(splits) > 1:
        return " ".join(splits[1:-1])
    else:
        raise ValueError("Unknown")


def cpubrand(x):
    splits = x.strip().split()
    if len(splits) > 1:
        return splits[0]
    else:
        raise ValueError("Unknown")


def cpupower(x):
    splits = x.strip().split()
    return float(splits[-1].replace("GHz", ""))


data["CpuBrand"] = data["Cpu"].apply(cpubrand)
data["CpuType"] = data["Cpu"].apply(cputype)
data["CpuPower"] = data["Cpu"].apply(cpupower)
data.rename(columns={"Cpu": "OldFeature_Cpu"}, inplace=True)

# print(data["CpuType"].unique())
# print(data["CpuBrand"].unique())
# print(data["CpuPower"].unique())


def ssd(x):
    splits = x.strip().split("+")
    for s in splits:
        if "SSD" in s:
            return float(s.split()[0].replace("GB", "").replace("TB", "000"))
    return 0


def hdd(x):
    splits = x.strip().split("+")
    for s in splits:
        if "HDD" in s:
            if "GB" in s:
                return float(s.split()[0].replace("GB", ""))
            elif "TB" in s:
                return float(s.split()[0].replace("TB", "")) * 1000
    return 0


def flash(x):
    splits = x.strip().split("+")
    for s in splits:
        if "Flash Storage" in s:
            if "GB" in s:
                return float(s.split()[0].replace("GB", ""))
            elif "TB" in s:
                return float(s.split()[0].replace("TB", "")) * 1000
    return 0


data["Memory_SSD (GB)"] = data["Memory"].apply(ssd)
data["Memory_HDD (GB)"] = data["Memory"].apply(hdd)
data["Memory_Flash (GB)"] = data["Memory"].apply(flash)
data.rename(columns={"Memory": "OldFeature_Memory"}, inplace=True)

# print(data["Memory_SSD (GB)"].unique())
# print(data["Memory_HDD (GB)"].unique())
# print(data["Memory_Flash (GB)"].unique())

data["Weight"] = data["Weight"].apply(lambda x: float(x.replace("kg", "")))
data["Price_euros"] = data["Price_euros"].apply(float)

print(data.columns)
useful_columns = [
    "Company",
    "TypeName",
    "Inches",
    "Ram",
    "OpSys",
    "Weight",
    "ScreenResolutionX",
    "ScreenResolutionY",
    "CpuBrand",
    "CpuPower",
    "Memory_SSD (GB)",
    "Memory_HDD (GB)",
    "Memory_Flash (GB)",
    "Price_euros",
]

data["Ram"] = data["Ram"].apply(lambda x: float(x.replace("GB", "")))
for col in useful_columns:
    print(col, data[col].unique() if len(data[col].unique()) < 20 else "")

data[useful_columns].to_csv(
    os.path.join(os.path.dirname(__file__), "laptops_clean.csv"), index=None
)



import torch
import gpytorch
from .base import BaseRegressor
from .gpytorch_models import ExactGPModel
import warnings


class ExactGPRegressor(BaseRegressor):
    """[summary]
    Call the constructor of base class after defining the model.
    """

    def __init__(self, train_x, train_y, kernel):
        super().__init__()
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        self.train_x = train_x
        self.train_y = train_y
        self.model = ExactGPModel(self.train_x, self.train_y, likelihood, kernel)
        self.mll = gpytorch.mlls.ExactMarginalLogLikelihood(
            self.model.likelihood, self.model
        )

    def forward(self, X, y):
        self.model.set_train_data(X, y, strict=False)
        output = self.model(X)
        return -self.mll(output, y)

    def predict(self, X_test):
        self.model.set_train_data(
            inputs=self.train_x, targets=self.train_y, strict=False
        )
        if len(self.history["train_loss"]) < 1:
            warnings.warn(
                "Model is not fitted yet. This may cause unexpected behavior."
            )

        self.model.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            self.pred_dist = self.model.likelihood(self.model(X_test))
            return self.pred_dist



import torch
import gpytorch


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood, kernel):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = kernel

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)



import torch
import gpytorch

# Test ExactGPRegressor
class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood, kernel):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = kernel

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class exact_gp_regressor_from_gpytorch:
    def __init__(self, data, kernel, seed, n_iters, device):
        # initialize likelihood and model
        torch.manual_seed(seed)
        self.likelihood = gpytorch.likelihoods.GaussianLikelihood()
        self.model = ExactGPModel(data.X_train, data.Y_train, self.likelihood, kernel)

        data.X_train = data.X_train.to(device)
        data.Y_train = data.Y_train.to(device)
        self.model = self.model.to(device)
        self.likelihood = self.likelihood.to(device)

        # Find optimal model hyperparameters
        self.model.train()
        self.likelihood.train()

        # Use the adam optimizer
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=0.1
        )  # Includes GaussianLikelihood parameters

        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood, self.model)

        for param in self.model.parameters():
            torch.nn.init.normal_(param, mean=0.0, std=1.0)
            # print("old", param)

        for _ in range(n_iters):
            # Zero gradients from previous iteration
            optimizer.zero_grad()
            # Output from model
            output = self.model(data.X_train)
            # Calc loss and backprop gradients
            loss = -mll(output, data.Y_train)
            loss.backward()
            optimizer.step()

        self.model.eval()
        self.likelihood.eval()

        # Test points are regularly spaced along [0,1]
        # Make predictions by feeding model through likelihood
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            self.pred_dist = self.likelihood(self.model(data.X_test.to(device)))



import os
import sys
import torch
from gpytorch.kernels import RBFKernel, ScaleKernel, IndexKernel
from ztest import HammingDistanceKernel
from skgpytorch.models import ExactGPRegressor
import matplotlib.pyplot as plt
from bayesian_benchmarks.data import Boston, Wilson_autompg
from sklearn.preprocessing import LabelEncoder, StandardScaler
from sklearn.model_selection import train_test_split
from skgpytorch.metrics import negative_log_predictive_density
import json
import pandas as pd


class Data:
    pass


if sys.argv[1] == "boston":
    data = Boston()
    cat_index = [3]

elif sys.argv[1] == "autompg":
    data = Wilson_autompg()
    cat_index = [0, 6]

elif sys.argv[1] == "laptops":
    laptops = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "laptops_clean.csv"), encoding="latin-1"
    )
    data = Data()
    data.name = "laptops"
    for i in range(laptops.shape[1] - 1):
        try:
            scaler = StandardScaler()
            laptops.iloc[:, i] = scaler.fit_transform(
                laptops.iloc[:, i].values.reshape(-1, 1)
            )
        except:
            scaler = LabelEncoder()
            laptops.iloc[:, i] = scaler.fit_transform(laptops.iloc[:, i].values)
            # print(laptops[laptops.columns[i]].unique())
            print("Labeling column {}".format(laptops.columns[i]))
    data.X = laptops.iloc[:, :-1].values
    data.y = laptops.iloc[:, -1].values
    data.y = StandardScaler().fit_transform(data.y.reshape(-1, 1)).ravel()
    data.X_train, data.X_test, data.Y_train, data.Y_test = train_test_split(
        data.X,
        data.y,
        test_size=0.2,
        random_state=42,
    )
    cat_index = [0, 1, 4, 8]
    encoder = LabelEncoder()
    for c in cat_index:
        encoder.fit(data.X[:, c])
        data.X_train[:, c] = encoder.transform(data.X_train[:, c])
        data.X_test[:, c] = encoder.transform(data.X_test[:, c])

if sys.argv[1] != "laptops":
    encoder = LabelEncoder()
    for c in cat_index:
        data.X_train[:, c] = encoder.fit_transform(data.X_train[:, c])
        data.X_test[:, c] = encoder.transform(data.X_test[:, c])

cont_index = sorted(set(range(data.X_train.shape[1])) - set(cat_index))
n_iters = 100
seed = 0


# print(set(data.X_train[:, 3]))
# sys.exit()

cont_kernel = RBFKernel(ard_num_dims=len(cont_index), active_dims=cont_index)
X = torch.tensor(data.X_train, dtype=torch.float)
Y = torch.tensor(data.Y_train, dtype=torch.float).ravel()

hamming_kernel = None
for c in cat_index:
    if hamming_kernel is None:
        hamming_kernel = HammingDistanceKernel(
            num_categories=len(set(data.X_train[:, c])),
            ard_num_dims=1,
            active_dims=[c],
        )
    else:
        hamming_kernel = hamming_kernel * HammingDistanceKernel(
            num_categories=len(set(data.X_train[:, c])),
            ard_num_dims=1,
            active_dims=[c],
        )

index_kernel = None
for c in cat_index:
    if index_kernel is None:
        index_kernel = IndexKernel(
            num_tasks=len(set(data.X_train[:, c])),
            ard_num_dims=1,
            active_dims=[c],
        )
    else:
        index_kernel = index_kernel * IndexKernel(
            num_tasks=len(set(data.X_train[:, c])),
            ard_num_dims=1,
            active_dims=[c],
        )

ii = 0
cat_kernels = [
    hamming_kernel,
    RBFKernel(ard_num_dims=len(cat_index), active_dims=cat_index),
    index_kernel,
][ii:]
cat_names = ["Hamming", "RBF", "Index"][ii:]
cat_colors = ["red", "green", "blue"][ii:]
titles = ["NLPD"]
nlpd = {i: [] for i in cat_names}
for i in range(len(cat_names)):
    for seed in range(20):
        print(i, seed)
        full_kernel = ScaleKernel(cont_kernel * cat_kernels[i])
        gp = ExactGPRegressor(
            X,
            Y,
            kernel=full_kernel,
            random_state=seed,
            verbose=0,
            device="cuda",
        )
        gp.fit(n_iters=n_iters)
        nlpd[cat_names[i]].append(
            negative_log_predictive_density(
                gp.model,
                gp.likelihood,
                torch.tensor(data.X_test, dtype=torch.float).cuda(),
                torch.tensor(data.Y_test, dtype=torch.float).cuda().ravel(),
            )
        )

        plt.plot(gp.history["train_loss"], color=cat_colors[i], linewidth=1)
    plt.plot(
        gp.history["train_loss"], label=cat_names[i], color=cat_colors[i], linewidth=1
    )
    tmp_title = (
        cat_names[i]
        + ": Mean = "
        + str(round(sum(nlpd[cat_names[i]]) / len(nlpd[cat_names[i]]), 4))
        + " Std: "
        + str(round(torch.std(torch.tensor(nlpd[cat_names[i]])).item(), 4))
    )
    print(tmp_title)
    titles.append(tmp_title)

plt.title("\n".join(titles))
plt.legend()
plt.tight_layout()
plt.savefig(os.path.join(os.path.dirname(__file__), f"loss_{data.name}.png"))
json.dump(
    nlpd, open(os.path.join(os.path.dirname(__file__), f"loss_{data.name}.json"), "w")
)



import torch
import gpytorch


def negative_log_predictive_density(dist, y):
    """
    Negative log predictive density of model (normalized by number of observations).
    """
    return -dist.log_prob(y.ravel().to(dist.mean.device)).item() / y.shape[0]


def mean_squared_error(dist, y, squared=True):
    """
    Mean Squared Error
    """
    mse = torch.square(y.ravel().to(dist.mean.device) - dist.mean).mean().item()
    if not squared:
        return mse ** 0.5  # Root mean square error
    return mse



import pytest
from bayesian_benchmarks.data import Boston

import torch
import gpytorch
from gpytorch.kernels import ScaleKernel, RBFKernel, MaternKernel
from skgpytorch.models import ExactGPRegressor
from .gpytorch_models import exact_gp_regressor_from_gpytorch
from skgpytorch.metrics import negative_log_predictive_density, mean_squared_error


@pytest.mark.comp_model
def test_exact_gp_regressor():
    device = "cuda" if torch.cuda.is_available() else "cpu"
    n_iters = 10

    data = Boston()
    data.X_train = torch.tensor(data.X_train, dtype=torch.float).to(device)
    data.Y_train = torch.tensor(data.Y_train, dtype=torch.float).ravel().to(device)
    data.X_test = torch.tensor(data.X_test, dtype=torch.float).to(device)
    data.Y_test = torch.tensor(data.Y_test, dtype=torch.float).ravel().to(device)
    kernels = [
        ScaleKernel(RBFKernel(ard_num_dims=data.X_train.shape[1])),
    ]

    seed = 0
    for kernel in kernels:
        gpytorch_model = exact_gp_regressor_from_gpytorch(
            data, kernel, seed, n_iters, device
        )
        gp = ExactGPRegressor(
            data.X_train,
            data.Y_train,
            kernel,
        )
        gp.to(device)
        gp.fit(n_epochs=n_iters, random_state=seed, batch_size=100)
        pred_dist = gp.predict(data.X_test)

        avg_var = (pred_dist.variance - gpytorch_model.pred_dist.variance).abs_().mean()
        avg_mean = (pred_dist.mean - gpytorch_model.pred_dist.mean).abs_().mean()
        print(avg_mean, avg_var)
        assert avg_var < 0.5
        assert avg_mean < 0.5

        rmse_a = mean_squared_error(pred_dist, data.Y_test, squared=False)
        rmse_b = mean_squared_error(
            gpytorch_model.pred_dist, data.Y_test, squared=False
        )

        assert abs(rmse_a - rmse_b) < 0.5

        nlpd_a = negative_log_predictive_density(pred_dist, data.Y_test)
        nlpd_b = negative_log_predictive_density(gpytorch_model.pred_dist, data.Y_test)
        assert abs(nlpd_a - nlpd_b) < 0.7



import torch
from .base import BaseRegressor
from nsgptorch.models import GP
import warnings


class ExactNSGPRegressor(BaseRegressor):
    """[summary]
    Call the constructor of base class after defining the model.
    """

    def __init__(self, kernel_list, input_dim, inducing_points, device="cpu"):
        super().__init__()
        self.device = device
        self.model = GP(kernel_list, input_dim, inducing_points)

    def forward(self, X, y):
        return self.model(X, y)

    def predict(self, X_orig, y_orig, X_test):
        return self.model.predict(
            X_orig.to(self.device), y_orig.to(self.device), X_test.to(self.device)
        )
